from typing import Callable, Mapping, Any

import inspect


def inject(fn: Callable, injections: Mapping[str, Any]):
    sig: inspect.Signature = inspect.signature(fn)
    parameters = sig.parameters.values()
    inject_p = {}

    for param in parameters:
        if param.name in injections:
            inject_p[param.name] = injections[param.name]
        if param.kind == param.VAR_KEYWORD:
            inject_p.update(injections)
            break

    return fn(**inject_p)



from .. import Magma, TokenAuth, Model
import pytest
import responses
from typing import Dict

@responses.activate
def test_magma_retrieve(labor_template: Dict):
    ''' it retrieves records from magma '''
    responses.add(
        responses.POST,
        "https://magma.test/retrieve",
        json={
			"models": {
				"labor": {
					"documents": {},
                    "template": labor_template
				}
			}
        },
        status=200
    )
    client = Magma(
        auth=TokenAuth(token='token'),
        hostname='magma.test'
    )
    response = client.retrieve('ipi')

    assert 'labor' in response.models
    assert type(response.models['labor']) == Model

@responses.activate
def test_magma_update(labor_template: Dict):
    ''' it updates records in magma '''
    responses.add(
        responses.POST,
        "https://magma.test/update",
        json={
			"models": {
				"labor": {
					"documents": {
                        "The Nemean Lion" : {
                            "country" : "Nemea"
                        }
                    },
                    "template": labor_template
				}
			}
        },
        status=200
    )
    client = Magma(
        auth=TokenAuth(token='token'),
        hostname='magma.test'
    )
    response = client.update({ "project_name":"ipi", "revisions": { "labor": { "The Nemean Lion" : { "country" : "Nemea" } } } } )
    assert response.models['labor'].documents['The Nemean Lion']['country'] == 'Nemea'

@responses.activate
def test_magma_query():
    ''' it queries magma '''
    responses.add(
        responses.POST,
        "https://magma.test/query",
        json={
            'answer': [ [ 'The Nemean Lion', 'Nemea' ] ],
            'format': [ 'labor#country' ]
        },
        status=200
    )
    client = Magma(
        auth=TokenAuth(token='token'),
        hostname='magma.test'
    )
    response = client.query({ "project_name":"ipi", "query": [ 'labor', [ 'country', '::equals', 'Nemea' ], '::all', 'country' ]})
    assert response.answer == [ [ 'The Nemean Lion', 'Nemea' ] ]


@responses.activate
def test_magma_query_tsv():
    ''' it queries magma '''
    body = "identifier\tcolumn1\tcolumn2\nrecord1\t1\tblahblah\nrecord2\t3.4\tanything you want"
    responses.add(
        responses.POST,
        "https://magma.test/query",
        body=body,
        status=200
    )
    client = Magma(
        auth=TokenAuth(token='token'),
        hostname='magma.test'
    )

    response = client.query({ "project_name":"ipi", "query": [ 'labor', [ 'country', '::equals', 'Nemea' ], '::all', 'country' ], "format": "tsv"})
    assert response.to_json() == '{"identifier":{"0":"record1","1":"record2"},"column1":{"0":1.0,"1":3.4},"column2":{"0":"blahblah","1":"anything you want"}}'


import io
from typing import Iterable


class IterStream(io.RawIOBase):
    def __init__(self, iterable):
        self.iterable = iterable
        self.leftover = None

    def readable(self):
        return True

    def readinto(self, b):
        try:
            l = len(b)  # We're supposed to return at most this much
            chunk = self.leftover or next(self.iterable)
            output, self.leftover = chunk[:l], chunk[l:]
            b[: len(output)] = output
            return len(output)
        except StopIteration:
            return 0  # indicate EOF


def iterable_to_stream(
    iterable: Iterable[bytes], buffer_size=io.DEFAULT_BUFFER_SIZE
) -> io.BufferedReader:
    """
    Lets you use an iterable (e.g. a generator) that yields bytestrings as a read-only
    input stream.

    The stream implements Python 3's newer I/O API (available in Python 2's io module).
    For efficiency, the stream is buffered.
    """
    return io.BufferedReader(IterStream(iterable), buffer_size=buffer_size)



from .. import Janus, EtnaSession, TokenAuth
import pytest
import responses
from typing import Dict

@responses.activate
def test_janus_generate_token():
    ''' it generates a task token via Janus '''
    responses.add(
        responses.POST,
        "https://janus.test/api/tokens/generate",
        body='task_token',
        status=200
    )
    client = Janus(
        auth=TokenAuth(token='token'),
        hostname='janus.test'
    )
    assert client.generate_token('labors') == 'task_token'



from .janus import Janus
from .metis import File, Folder, Metis, TailResultContainer, TailNode
from .magma import Magma, Model, Template, Attribute, UpdateRequest, QueryRequest
from .polyphemus import Polyphemus
from .gnomon import Gnomon
from .auth import TokenAuth, SigAuth
from .etna_base import EtnaSession
from . import janus
from . import metis
from . import magma
from . import polyphemus
from . import gnomon
from . import auth
from . import etna_base

from . import utils



from dataclasses import dataclass, field
import typing
from datetime import datetime, timedelta
from io import StringIO
from functools import partial
from inspect import isgenerator
from typing import Dict, Optional, List
from serde import serialize, deserialize
from serde.json import from_json

import pandas as pd

from .etna_base import EtnaClientBase
from requests import HTTPError
from .utils.iterables import batch_iterable
from .metis import Upload, File

@serialize
@deserialize
@dataclass
class EtlConfigResponse:
    ran_at: Optional[str] = None
    status: str = ""
    created_at: str = ""
    comment: str = ""
    project_name: str = ""
    name: str = ""
    etl: str = ""
    config_id: int = 0
    version_number: int = 0
    config: Dict = field(default_factory=dict)
    secrets: Dict = field(default_factory=dict)
    params: Dict = field(default_factory=dict)
    run_interval: int = -1

    def should_run(self, start, end):
        # run_never and run_archive
        if self.run_interval < 0:
            return False

        # run_once always run
        if self.run_interval == 0:
            return True

        # new intervals always run
        if not self.ran_at:
            return True

        ran_at = datetime.fromisoformat(self.ran_at)

        if ran_at + timedelta(0, self.run_interval) < start:
            return True

        return False

@serialize
@deserialize
@dataclass
class ListEtlConfigsResponse:
    configs: List[EtlConfigResponse] = field(default_factory=list)

class Polyphemus(EtnaClientBase):
    def list_all_etl_configs(
        self, job_type: Optional[str] = None
    ):
        args = dict()

        if job_type is not None:
            args["job_type"] = job_type
        response = self.session.post(
            self.prepare_url("api", "etl", "configs"),
            json=args
        )

        return from_json(ListEtlConfigsResponse, response.content)

    def etl_update(
        self,
        project_name: str,
        config_id: str,
        **update
    ):
        response = self.session.post(
            self.prepare_url("api", "etl", project_name, "update", str(config_id)),
            json=update
        )

        return from_json(EtlConfigResponse, response.content)

    def etl_add_output(
        self,
        project_name: str,
        config_id: str,
        append=False,
        output: str =""
    ):
        response = self.session.post(
            self.prepare_url("api", "etl", project_name, "output", str(config_id)),
            json=dict(append=append, output=output)
        )

        return from_json(EtlConfigResponse, response.content)



from .etna_base import EtnaClientBase

class Janus(EtnaClientBase):
    def generate_token(
        self, project_name: str, read_only=True, token_type="task"
    ) -> str:
        response = self.session.post(
            self.prepare_url("api", "tokens", "generate"),
            json=dict(
                token_type=token_type, project_name=project_name, read_only=read_only
            ),
        )
        return response.content.decode(encoding='utf-8')

    def nonce(self) -> bytes:
        response = self.session.get(self.prepare_url("api", "tokens", "nonce"))
        return response.content.decode(encoding='utf-8')



import requests
from typing import Optional
from requests.auth import AuthBase
from cryptography.hazmat.primitives import serialization, hashes
from cryptography.hazmat.primitives.asymmetric import padding
from requests.auth import AuthBase
import base64

class TokenAuth(AuthBase):
    token: str
    project_scope: Optional[str]

    def __init__(self, token: str, project_scope: Optional[str] = None):
        self.token = token
        self.project_scope = project_scope

    def __call__(self, r: requests.Request) -> requests.Request:
        r.headers["Authorization"] = f'Etna {self.token}'
        return r

class SigAuth(AuthBase):
    nonce: str
    email: str
    private_key: bytes

    def __init__(self, private_key_file: str, email: str, nonce: str):
        self.private_key = open(private_key_file, "rb").read()
        self.nonce = nonce
        self.email = email

    def __call__(self, r: requests.Request) -> requests.Request:
        key = serialization.load_pem_private_key(self.private_key, None)
        txt_to_sign: bytes = b".".join(
            [self.nonce.encode('ascii'), base64.b64encode(self.email.encode("ascii"))]
        )
        sig = base64.b64encode(
            key.sign(
                txt_to_sign,
                padding.PKCS1v15(
                    # If in the future the server changes to PSS
                    # mgf=padding.MGF1(hashes.SHA256()),
                    # salt_length=padding.PSS.MAX_LENGTH
                ),
                hashes.SHA256(),
            )
        )
        together: bytes = b".".join([txt_to_sign, sig])
        r.headers["Authorization"] = f'Signed-Nonce {together.decode("ascii")}'
        return r




import pytest

@pytest.fixture
def labor_template():
    return {
        "name": "labor",
        "attributes": {
            "created_at": {
                "name": "created_at",
                "attribute_name": "created_at",
                "display_name": "Created At",
                "hidden": True,
                "validation": None,
                "attribute_type": "date_time"
            },
            "updated_at": {
                "name": "updated_at",
                "attribute_name": "updated_at",
                "display_name": "Updated At",
                "hidden": True,
                "validation": None,
                "attribute_type": "date_time"
            },
            "name": {
                "name": "name",
                "attribute_name": "name",
                "description": "Name for this labor.",
                "display_name": "Name",
                "restricted": False,
                "read_only": False,
                "hidden": False,
                "validation": None,
                "attribute_type": "identifier"
            },
            "country": {
                "name": "country",
                "attribute_name": "country",
                "description": "Place where this labor took place.",
                "restricted": False,
                "read_only": False,
                "hidden": False,
                "validation": None,
                "attribute_type": "string"
            }
        },
        "identifier": "name"
    }



from typing import Tuple, Generic, Callable, TypeVar, Any, List, Union
import re

CursorIdx = int
A = TypeVar("A")
B = TypeVar("B")


class ParseFailure(Exception):
    pass


class Parser(Generic[A]):
    parse: Callable[[str, CursorIdx], Tuple[A, CursorIdx]]

    def __init__(self, p: Callable[[str, CursorIdx], Tuple[A, CursorIdx]]):
        self.parse = p

    def run_parser(self, s: str) -> Union[Tuple[A, None], Tuple[None, ParseFailure]]:
        try:
            result = self.parse(s, 0)[0]
            return result, None
        except ParseFailure as e:
            return None, e


# This is the Functor of Parser that simply transform the result
# of any parse on the given parser by the given callable.
def parser_map(a: "Parser[A]", f: Callable[[A], B]) -> "Parser[B]":
    def map_result(s: str, c: CursorIdx):
        v, c = a.parse(s, c)
        return f(v), c

    return Parser(map_result)


def parser_lift(v: A) -> "Parser[A]":
    return Parser(lambda s, c: (v, c))


def parser_apply(f: "Parser[Callable[[A], B]]", a: "Parser[A]") -> "Parser[B]":
    def apply_result(s: str, c: CursorIdx) -> Tuple[B, CursorIdx]:
        next_work, c = f.parse(s, c)
        next_a, c = a.parse(s, c)
        return next_work(next_a), c

    return Parser(apply_result)


def parser_bind(f: "Parser[Callable[[A], Parser[B]]]", a: "Parser[A]") -> "Parser[B]":
    def bind_result(s: str, c: CursorIdx) -> Tuple[B, CursorIdx]:
        next_work, c = f.parse(s, c)
        next_a, c = a.parse(s, c)
        next_parser: "Parser[B]" = next_work(next_a)
        return next_parser.parse(s, c)

    return Parser(bind_result)


def either_parser(a: Parser[A], *alternatives: Parser[A]) -> Parser[A]:
    def try_one_then(s: str, c: CursorIdx) -> Tuple[A, CursorIdx]:
        try:
            return a.parse(s, c)
        except ParseFailure as e:
            last_exception = e

        for alt in alternatives:
            try:
                return alt.parse(s, c)
            except ParseFailure as e:
                last_exception = e

        raise last_exception

    return Parser(try_one_then)


def parse_many(parser: Parser[A]) -> Parser[List[A]]:
    def do_parsing(input_text: str, cursor: CursorIdx):
        items: List[A] = []

        while True:
            try:
                result, new_cursor = parser.parse(input_text, cursor)
                if new_cursor <= cursor:
                    raise ParseFailure(
                        "Potential infinite recursion in parse_many, cursor did not move forward."
                    )
                cursor = new_cursor
                items.append(result)
            except ParseFailure as failure:
                break

        return items, cursor

    return Parser(do_parsing)


# "Lift" a parsing value into a cell (singleton array) if it succeeds,
# otherwise provide an empty array.
def parser_maybe(parser: Parser[A]) -> Parser[List[A]]:
    return either_parser(parser_map(parser, lambda v: [v]), parser_lift([]))


def parse_many_joined_by(parser: Parser[A], sep: Parser[Any]) -> Parser[List[A]]:
    def with_head(head: List[A]) -> Parser[List[A]]:
        return parser_map(parser_maybe(parser), lambda tail: head + tail)

    each_head: Parser[A] = take_left(parser, sep)
    head_parser: Parser[List[A]] = parse_many(each_head)
    return parser_bind(parser_lift(with_head), head_parser)


def regex_parser(pattern: re.Pattern) -> Parser[str]:
    def do_parsing(input_text: str, cursor: CursorIdx):
        m = pattern.match(input_text, cursor)
        if m is None:
            raise ParseFailure(f"No match found for {pattern} at position {cursor}")

        start, end = m.span()
        return input_text[start:end], end

    return Parser(do_parsing)


def lift_parse_failure(message: str) -> Parser[Any]:
    def do_parsing(input_text: str, cursor: CursorIdx):
        raise ParseFailure(message)

    return Parser(do_parsing)


def literal_parser(literal: str) -> Parser[str]:
    def do_parsing(input_text: str, cursor: CursorIdx):
        next = input_text[cursor : cursor + len(literal)]
        if next != literal:
            raise ParseFailure(f"No match found for '{literal}' at position {cursor}")

        return literal, cursor + len(literal)

    return Parser(do_parsing)


def take_right(l: Parser[Any], r: Parser[A]) -> Parser[A]:
    def compose(_):
        def get_value(v):
            return v

        return get_value

    return parser_apply(parser_apply(parser_lift(compose), l), r)


def take_left(l: Parser[A], r: Parser[Any]) -> Parser[A]:
    def compose(v):
        def skip(_):
            return v

        return skip

    return parser_apply(parser_apply(parser_lift(compose), l), r)


def surrounded_by(l: Parser[Any], p: Parser[A], r: Parser[Any]) -> Parser[A]:
    return take_left(take_right(l, p), r)


whitespace_skipper: Parser[str] = regex_parser(re.compile(r"\s*"))


def token(p: Parser[A], bounds: Parser[Any] = whitespace_skipper) -> Parser[A]:
    return surrounded_by(bounds, p, bounds)


int_parser: Parser[int] = parser_map(regex_parser(re.compile(r"[-+]?[0-9]+")), int)
float_parser: Parser[float] = parser_map(regex_parser(re.compile(r"0\.\d\d+")), float)
eof_parser: Parser[str] = regex_parser(re.compile(r"$"))



import time


class RetryGater:
    def __init__(self, max: int, f=1, exp=2):
        self.max = max
        self.i = 0
        self.last = 0
        self.exp = exp
        self.f = f

    def reset(self):
        self.i = 0

    def gate(self, exception: Exception):
        self.i += 1
        if self.i > self.max:
            raise exception

        time.sleep((self.exp ** (self.i - 1)) * self.f)



import dateutil
from datetime import datetime
from typing import Dict, Optional, List
import typing
import hashlib
import io
import re
import os.path
import uuid
import dataclasses
from serde import serialize, deserialize
from serde.json import from_json, to_json
from .etna_base import EtnaClientBase
from .utils.multipart import encode_as_multipart
from .utils.streaming import iterable_to_stream
from requests import HTTPError

@serialize
@deserialize
@dataclasses.dataclass
class MagmaFileEntry:
    path: str = ""
    original_filename: str = ""

@serialize
@deserialize
@dataclasses.dataclass
class TailNode:
    id: int = 0
    type: str = "file"
    node_name: str = ""
    updated_at: str = ""
    parent_id: Optional[int] = None
    file_hash: Optional[str] = None
    archive_id: Optional[int] = None

    @property
    def updated_at_datetime(self) -> datetime:
        return dateutil.parser.parse(self.updated_at)

class TailResultContainer:
    bucket_name: str
    project_name: str
    nodes: typing.List[TailNode]
    parents: typing.Dict[int, TailNode]
    path_cache: typing.Dict[int, str]

    def __init__(self, bucket_name: str, project_name: str):
        self.nodes = []
        self.parents = {}
        self.bucket_name = bucket_name
        self.project_name = project_name
        self.path_cache = {}

    def add(self, node: TailNode):
        if node.type == 'parent':
            self.parents[node.id] = node
        else:
            self.nodes.append(node)

    def resolve_files(self) -> List["File"]:
        result = []
        for node in self.nodes:
            if node.type != 'file':
                continue
            result.append(File(
                file_name=node.node_name,
                file_hash=node.file_hash,
                updated_at=node.updated_at,
                file_path=self.resolve_path(node),
                folder_id=node.parent_id,
                project_name=self.project_name,
                bucket_name=self.bucket_name,
            ))

        return result

    def resolve_folders(self, include_parents=False) -> List["Folder"]:
        result = []
        for node in self.nodes:
            if node.type != 'folder':
                continue
            result.append(self._construct_folder(node))

        if include_parents:
            result.extend(self.resolve_parents())

        return result

    def resolve_parents(self) -> List["Folder"]:
        result = []
        for node in self.parents.values():
            result.append(self._construct_folder(node))

        return result

    def _construct_folder(self, node: TailNode):
        return Folder(
            folder_name=node.node_name,
            updated_at=node.updated_at,
            folder_path=self.resolve_path(node),
            id=node.id,
            project_name=self.project_name,
            bucket_name=self.bucket_name,
            folder_id=node.parent_id,
        )

    def resolve_path(self, node: TailNode):
        if node.parent_id is None:
            return node.node_name

        parent_path = self.path_cache.get(node.parent_id)

        if parent_path is None:
            parent = self.parents[node.parent_id]
            parent_path = self.resolve_path(parent)
            self.path_cache[node.parent_id] = parent_path

        return f"{parent_path}/{node.node_name}"

@serialize
@deserialize
@dataclasses.dataclass
class File:
    file_name: str = ""
    file_hash: str = ""
    updated_at: str = ""
    file_path: Optional[str] = ""
    folder_id: Optional[int] = None
    project_name: str = ""
    bucket_name: str = ""
    download_url: Optional[str] = ""

    def __str__(self):
        return self.as_metis_url

    @property
    def as_magma_file_attribute(self) -> MagmaFileEntry:
        return MagmaFileEntry(
            path=self.as_metis_url,
            original_filename=self.file_name or os.path.basename(self.file_path),
        )

    @property
    def as_metis_url(self):
        return f"metis://{self.project_name}/{self.bucket_name}/{self.file_path}"

    @property
    def updated_at_datetime(self) -> datetime:
        return dateutil.parser.parse(self.updated_at)


@serialize
@deserialize
@dataclasses.dataclass
class Folder:
    id: int = 0
    folder_path: str = ""
    project_name: str = ""
    bucket_name: str = ""
    folder_name: str = ""
    updated_at: str = ""
    folder_id: int = 0

    @property
    def updated_at_datetime(self) -> datetime:
        return dateutil.parser.parse(self.updated_at)


@serialize
@deserialize
@dataclasses.dataclass
class FoldersAndFilesResponse:
    folders: List[Folder] = dataclasses.field(default_factory=list)
    files: List[File] = dataclasses.field(default_factory=list)

    def empty(self):
        return not self.folders and not self.files

    def extend(self, other: "FoldersAndFilesResponse"):
        self.folders.extend(other.folders)
        self.files.extend(other.files)


@serialize
@deserialize
@dataclasses.dataclass
class FoldersResponse:
    folders: List[Folder]


@serialize
@deserialize
@dataclasses.dataclass
class UploadResponse:
    current_byte_position: int = 0
    url: str = ""
    next_blob_size: int = 0

    @property
    def upload_path(self):
        return re.compile(r"^https://[^/]*?/").sub("/", self.url)


@dataclasses.dataclass
class Upload:
    file: typing.IO
    file_size: int
    upload_path: str

    cur: int = 0

    last_bytes: Optional[bytes] = None
    read_position: int = 0

    INITIAL_BLOB_SIZE = 2 ** 10
    MAX_BLOB_SIZE = 2 ** 22
    ZERO_HASH = "d41d8cd98f00b204e9800998ecf8427e"

    @property
    def as_magma_file_attribute(self) -> Optional[MagmaFileEntry]:
        metis_url = self.as_metis_url
        if metis_url:
            return MagmaFileEntry(
                path=metis_url, original_filename=os.path.basename(metis_url)
            )
        return None

    UPLOAD_PATH_REGEX = re.compile(r"/([^/]+)/upload/([^/]+)/?([^?]*)")

    @property
    def as_metis_url(self) -> Optional[str]:
        match = self.UPLOAD_PATH_REGEX.match(self.upload_path)
        if match:
            project_name = match.group(1)
            bucket_name = match.group(2)
            path = match.group(3)
            return f"metis://{project_name}/{bucket_name}/{path}"
        return None

    @property
    def as_file(self) -> Optional[File]:
        match = self.UPLOAD_PATH_REGEX.match(self.upload_path)
        if match:
            project_name = match.group(1)
            bucket_name = match.group(2)
            path = match.group(3)
            return File(file_name=os.path.basename(path), file_path=path, project_name=project_name, bucket_name=bucket_name)
        return None

    @property
    def next_blob_size(self):
        return min(
            self.MAX_BLOB_SIZE if self.cur > 0 else self.INITIAL_BLOB_SIZE,
            self.file_size - self.cur,
        )

    @property
    def is_complete(self):
        return self.cur >= self.file_size

    @property
    def next_blob_hash(self):
        b = self.read_next_blob()
        if not b:
            return self.ZERO_HASH

        return hashlib.md5(b).hexdigest()

    def resume_from(self, upload_response: UploadResponse):
        self.cur = upload_response.current_byte_position

    def advance_position(self):
        self.cur += self.next_blob_size

    def read_next_blob(self) -> bytes:
        next_left = self.cur
        next_right = self.cur + self.next_blob_size

        # Advance the position, if we've resumed to a specific place that needs skipping to.
        # We are merely reading data to catch up to our expected left position.
        if next_right < self.read_position:
            if self.file.seekable():
                self.file.seek(next_left)
                self.read_position = next_left
            else:
                raise ValueError(
                    "Upload source stream requires restart, upload has failed."
                )
        elif self.read_position < next_left:
            bytes_to_read = next_left - self.read_position
            data = self.file.read(bytes_to_read)
            self.read_position += len(data)
            if not len(data) < bytes_to_read:
                raise ValueError("Unexpected EOF while reading source stream.")

        if next_right == self.read_position:
            return self.last_bytes or b""

        if next_left != self.read_position:
            if self.file.seekable():
                self.file.seek(self.cur)
                self.read_position = self.cur
            else:
                raise ValueError(
                    "Alignment error, upload needs to restart but source is not seekable."
                )

        data: List[bytes] = []
        while self.read_position < next_right:
            expected_bytes = next_right - self.read_position
            chunk = self.file.read(expected_bytes)
            self.read_position += len(chunk)
            if len(chunk) < expected_bytes:
                raise ValueError("Unexpected EOF while reading source stream")
            data.append(chunk)

        self.last_bytes = b"".join(data)
        return self.last_bytes
    
    def reset(self):
        self.read_position = 0
        self.last_bytes = None


@serialize
@deserialize
@dataclasses.dataclass
class UploadStartRequest:
    file_size: int = 0
    action: str = "start"
    metis_uid: str = ""
    next_blob_size: int = 0
    upload_path: str = ""
    next_blob_hash: str = ""
    reset: bool = False


@dataclasses.dataclass
class UploadBlobRequest:
    file_size: int = 0
    action: str = "blob"
    metis_uid: str = ""
    blob_data: bytes = b""
    upload_path: str = ""
    next_blob_size: int = 0
    next_blob_hash: str = ""
    current_byte_position: int = 0

    def to_dict(self):
        return self.__dict__



class Metis(EtnaClientBase):
    def touch_folder(
        self, project_name: str, bucket_name: str, folder_path: str
    ) -> FoldersResponse:
        response = self.session.get(
            self.prepare_url(project_name, "folder", "touch", bucket_name, folder_path)
        )
        return from_json(FoldersResponse, response.content)

    def list_folder(
        self, project_name: str, bucket_name: str, folder_path: Optional[str] = None
    ) -> FoldersAndFilesResponse:
        if folder_path:
            response = self.session.get(
                self.prepare_url(project_name, "list", bucket_name, folder_path)
            )
        else:
            response = self.session.get(
                self.prepare_url(project_name, "list", bucket_name)
            )
        return from_json(FoldersAndFilesResponse, response.content)

    def authorize_download(self, project_name: str, bucket_name: str, file_path: str) -> str:
        response = self.session.post(
            self.prepare_url('authorize', 'download'),
            json=dict(
                project_name=project_name,
                bucket_name=bucket_name,
                file_path=file_path,
            )
        )

        return response.json()['download_url']

    def open_file(self, file: File, binary_mode=False) -> io.BufferedReader:
        """
        See open_url, this method opens the given file's download_url using that method.
        If the given file does not have a download_url set, it will fetch one using `authorize_download`.
        """
        download_url = file.download_url
        if not download_url:
            download_url = self.authorize_download(file.project_name, file.bucket_name, file.file_path)

        return self.open_url(download_url, binary_mode)

    def open_url(self, url: str, binary_mode=False) -> io.BufferedReader:
        """
        Opens the given url for download into a context as a python io (file like) object.
        By default, when binary_mode is False, the underlying io object yields decoded strings using utf-8.
        Otherwise, the provided io object yields bytes objects.

        Note:  Ideally, this method is used in combination with 'with' syntax in python, so that the underlying
        data stream is closed after usage.  This is especially performant when code only needs to access a small
        subset of the data.

        eg:
        ```
        with metis.open_url(file.download_url) as open_file:
          for line in csv.reader(open_file):
             break
        ```
        """
        response = self.session.get(url, stream=True)
        io_obj = iterable_to_stream(response.iter_content(io.DEFAULT_BUFFER_SIZE))
        if not binary_mode:
            io_obj = io.TextIOWrapper(io_obj, encoding='utf8')
        return io_obj

    def authorize_upload(
        self, project_name: str, bucket_name: str, file_path: str
    ) -> "UploadResponse":
        response = self.session.post(
            self.prepare_url("authorize", "upload"),
            json=dict(
                project_name=project_name, bucket_name=bucket_name, file_path=file_path
            ),
        )
        response_obj = from_json(UploadResponse, response.content)
        return response_obj

    def start_upload(self, request: "UploadStartRequest") -> "UploadResponse":
        response = self.session.post(
            self.prepare_url_unsafe(request.upload_path),
            data=to_json(request),
            headers={"Content-Type": "application/json"},
        )
        response_obj = from_json(UploadResponse, response.content)
        return response_obj

    def upload_blob(self, request: "UploadBlobRequest"):
        response = self.session.post(
            self.prepare_url_unsafe(request.upload_path),
            files=encode_as_multipart(request),
        )
        response_obj = from_json(UploadResponse, response.content)
        return response_obj

    def upload_file(
        self,
        project_name: str,
        bucket_name: str,
        dest_path: str,
        file: typing.IO,
        size: Optional[int] = None,
        max_retries=3,
    ) -> typing.Iterable["Upload"]:
        if size is None:
            if file.seekable():
                file.seek(0, 2)
                size = file.tell()
                file.seek(0)
            else:
                raise ValueError(
                    "size was not specified, but file is streaming.  You must specify a size hint when using a stream."
                )

        if len(dest_path) > 1 and not self.folder_exists(project_name, bucket_name, os.path.dirname(dest_path)):
            self.create_folder(project_name, bucket_name, os.path.dirname(dest_path))
        authorization = self.authorize_upload(project_name, bucket_name, dest_path)
        upload = Upload(file=file, file_size=size, upload_path=authorization.upload_path)

        return self.upload_parts(upload, uuid.uuid4().hex, max_retries)

    def upload_parts(
        self, upload: "Upload", metis_uid: str, remaining_attempts: int, reset=False
    ) -> typing.Iterable["Upload"]:
        """
        Note that the remaining_attempts here covers a specific use-case,
            where the upload does not match what has been uploaded to
            the server, and it needs to be reset / start over.
            It should not be used as a counter for retrying individual requests.
        """
        unsent_zero_byte_file = upload.cur == 0
        upload.resume_from(
            self.start_upload(
                UploadStartRequest(
                    upload_path=upload.upload_path,
                    file_size=upload.file_size,
                    next_blob_size=upload.next_blob_size,
                    next_blob_hash=upload.next_blob_hash,
                    metis_uid=metis_uid,
                    reset=reset,
                )
            )
        )

        if reset:
            upload.reset()

        while unsent_zero_byte_file or not upload.is_complete:
            try:
                current_byte_position = upload.cur
                blob_data = upload.read_next_blob()

                upload.advance_position()

                self.upload_blob(
                    UploadBlobRequest(
                        upload_path=upload.upload_path,
                        next_blob_size=upload.next_blob_size,
                        next_blob_hash=upload.next_blob_hash,
                        blob_data=blob_data,
                        metis_uid=metis_uid,
                        current_byte_position=current_byte_position,
                    )
                )

                unsent_zero_byte_file = False
            except HTTPError as e:
                if remaining_attempts > 1 and e.response.status_code == 422:
                    yield from self.upload_parts(
                        upload, metis_uid, remaining_attempts - 1, True
                    )
                    return
                else:
                    raise e

            yield upload

    def create_folder(
        self, project_name: str, bucket_name: str, folder_path: str
    ) -> FoldersAndFilesResponse:
        response = self.session.post(
            self.prepare_url(project_name, "folder", "create", bucket_name, folder_path)
        )
        response_obj = from_json(FoldersAndFilesResponse, response.content)
        return response_obj

    def folder_exists(
        self, project_name: str, bucket_name: str, folder_path: str
    ) -> bool:
        try:
            response = self.session.get(
                self.prepare_url(project_name, "list", bucket_name, folder_path)
            )
            return True
        except HTTPError as e:
            if e.response.status_code == 422:
                return False
            raise e

    def tail(
            self,
            project_name: str,
            bucket_name: str,
            type: typing.Union[typing.Literal['folders'], typing.Literal['files']],
            batch_start: Optional[datetime] = None,
            batch_end: Optional[datetime] = None,
            folder_id: Optional[typing.Union[int, typing.List[int]]] = None,
            include_parents: Optional[bool] = False,
     ) -> typing.Tuple[List[File], List[Folder]]:
        container = TailResultContainer(bucket_name, project_name)
        if batch_start and batch_end:
            args = dict(
                batch_start=batch_start.isoformat(timespec="seconds"),
                batch_end=batch_end.isoformat(timespec="seconds"),
                type=type,
            )
        else:
            args = dict(
                folder_id=folder_id,
                type=type,
            )

        response = self.session.post(
            self.prepare_url(project_name, 'tail', bucket_name),
            json=args, stream=True
        )
        for line in response.iter_lines():
            if line:
                container.add(from_json(TailNode, line))

        return container.resolve_files(), container.resolve_folders(include_parents=include_parents)

    def find(
        self,
        project_name: str,
        bucket_name: str,
        params: List[typing.Mapping],
        limit: Optional[int] = None,
        offset: Optional[int] = None,
        hide_paths=False,
    ) -> FoldersAndFilesResponse:
        limit_set = True

        if not limit:
            limit_set = False
            limit = 1000
        if not offset:
            offset = 0

        args = dict(
            params=params,
            hide_paths=hide_paths,
            limit=limit,
            offset=offset,
        )

        result = FoldersAndFilesResponse(folders=[], files=[])
        while True:
            response = self.session.post(
                self.prepare_url(project_name, "find", bucket_name), json=args
            )
            response_obj = from_json(FoldersAndFilesResponse, response.content)
            result.extend(response_obj)

            if response_obj.empty() or limit_set:
                break

            args["offset"] += max(len(response_obj.files), len(response_obj.folders))

        return result



from .. import Polyphemus, EtnaSession, TokenAuth
from ..polyphemus import ListEtlConfigsResponse
import pytest
import responses
from typing import Dict
import json
from datetime import datetime, timedelta

@responses.activate
def test_polyphemus_etl_config_should_run():
    ''' it should run in the indicated dates'''
    ran_at = datetime.now() - timedelta(1)
    start = datetime.now()
    end = datetime.now() + timedelta(1)

    configs = {
        'configs': [
            {
                'project_name': 'labors',
                'ran_at': (ran_at).isoformat(),
                'run_interval': 600
            }
        ]
    }
    responses.add(
        responses.POST,
        "https://polyphemus.test/api/etl/configs",
        body=json.dumps(configs),
        status=200
    )
    client = Polyphemus(
        auth=TokenAuth(token='token'),
        hostname='polyphemus.test'
    )
    response = client.list_all_etl_configs()
    config = response.configs[0]

    assert config.should_run(start, end)
    assert not config.should_run(ran_at, end)



from typing import Iterable, TypeVar, List

T = TypeVar("T")


def batch_iterable(iter: Iterable[T], batch_size: int) -> Iterable[Iterable[T]]:
    cur_batch: List[T] = []
    for item in iter:
        cur_batch.append(item)
        if len(cur_batch) >= batch_size:
            yield cur_batch
            cur_batch = []
    if cur_batch:
        yield cur_batch



from .. import Metis, TokenAuth
import responses
from responses import GET, POST, add
from io import BytesIO

@responses.activate
def test_metis_upload_file():
    ''' it sends a file to metis '''
    add(GET, "https://metis.test/athena/list/plans/helmet", json={}, status=422)
    add(POST, "https://metis.test/athena/folder/create/plans/helmet", json={}, status=200)
    add(POST, "https://metis.test/authorize/upload", json={'url': 'upload/plans/helmet/helmet.txt'}, status=200)
    add(POST, "https://metis.test/upload/plans/helmet/helmet.txt", json={}, status=200)
    client = Metis(
        auth=TokenAuth(token='token'),
        hostname='metis.test'
    )
    helmet_file = BytesIO(b'''
      xXx
     xO|Ox
    ''')
    for upload in client.upload_file('athena', 'plans', 'helmet/helmet.txt', helmet_file):
        print('Uploading')



from typing import Any, Dict, Iterable, Tuple, Union
from serde import to_dict


def encode_as_multipart(v: Any) -> Dict:
    """
    Can be passed as the files= parameter of a requests.Session post call, for instance.
    """
    result = {}

    if not isinstance(v, dict):
        if hasattr(v, 'to_dict'):
            v = v.to_dict()
        else:
            v = to_dict(v)

    for k, v in _encode_as_multipart(v, "", True):
        result[k] = v

    return result


def _encode_as_multipart(
    v: Any, base_key: str, is_root: bool = False
) -> Iterable[Tuple[str, Union[str, Tuple[str, bytes, str]]]]:
    if isinstance(v, dict):
        for k, v in v.items():
            yield from _encode_as_multipart(v, k if is_root else f"{base_key}[{k}]")
    elif isinstance(v, list):
        for i, item in enumerate(v):
            if isinstance(v, dict):
                # This is necessary to ensure that arrays of hashes that have hetergenous keys still get parsed correctly
                # Since the only way to indicate a new entry in the array of hashes is by re-using a key that existed in
                # the previous hash.
                yield from _encode_as_multipart(i, f"{base_key}[][_idx]")

            yield from _encode_as_multipart(v, f"{base_key}[]")
    else:
        if not base_key:
            raise ValueError("base_key cannot be empty for scalars!")

        if isinstance(v, bytes):
            yield base_key, ("blob", v, "application/octet-stream")
        else:
            yield base_key, (None, str(v))



import dataclasses
from typing import Dict, Optional, List
from serde import serialize, deserialize
from serde.json import from_json, to_json

from .etna_base import EtnaClientBase

@serialize
@deserialize
@dataclasses.dataclass
class RulesResponse:
    rules: Dict[ str, Dict[str, str ] ] = dataclasses.field(default_factory=dict)

class Gnomon(EtnaClientBase):
    def rules(
        self, project_names: List[str] = dataclasses.field(default_factory=list)
    ):
        response = self.session.post(
            self.prepare_url("gnomon", "rules"),
            json=dict(project_names=project_names)
        )

        return from_json(RulesResponse, response.content)



import dataclasses
import typing
from io import StringIO
from functools import partial
from inspect import isgenerator
from typing import Dict, Optional, List
from serde import serialize, deserialize
from serde.json import from_json, to_json

import pandas as pd

from .etna_base import EtnaClientBase
from requests import HTTPError
from .utils.iterables import batch_iterable
from .metis import Upload, File


@serialize
@deserialize
@dataclasses.dataclass
class Attribute:
    attribute_type: str = ""
    link_model_name: Optional[str] = None
    match: Optional[str] = None
    restricted: Optional[bool] = None
    read_only: Optional[bool] = None
    hidden: Optional[bool] = None
    validation: typing.Any = None


@serialize
@deserialize
@dataclasses.dataclass
class Template:
    attributes: Dict[str, Attribute]
    name: str = ""
    identifier: str = ""
    version: int = 0
    parent: str = ""


@serialize
@deserialize
@dataclasses.dataclass
class Model:
    documents: Dict[str, Dict[str, typing.Any]] = dataclasses.field(default_factory=dict)
    template: Optional[Template] = None
    count: int = 0

    def empty(self):
        return len(self.documents) == 0

    def extend(self, other: "Model"):
        if other.template:
            self.template = other.template
        self.count = other.count
        self.documents.update(other.documents)


@serialize
@deserialize
@dataclasses.dataclass
class UpdateRequest:
    revisions: Dict[str, Dict[str, Dict[str, typing.Any]]] = dataclasses.field(
        default_factory=dict
    )
    project_name: str = ""
    dry_run: bool = False

    def __bool__(self):
        return not self.empty()

    def shallow_copy(self, revisions: Optional[Dict[str, Dict[str, Dict[str, typing.Any]]]] = None) -> "UpdateRequest":
        if revisions is None:
            revisions = self.revisions

        return UpdateRequest(project_name=self.project_name, dry_run=self.dry_run, revisions=dict(**revisions))

    def includes(self, model_name: str, record_name: str):
        return record_name in self.revisions.get(model_name, {})

    def __iter__(self) -> typing.Iterator[typing.Tuple[str, str]]:
        for model_name, revisions in self.revisions.items():
            for record_name in revisions.keys():
                yield model_name, record_name

    def sample_revision_tree(self, model_name: str, record_name: str, models: Dict[str, Model], acc: "UpdateRequest", destructive=False):
        """
        Attempts to sample the complete graph of related objects reference by the given model_name / record_name in this
        update, copying it and all its referenced links to acc.

        When destructive=True is set, it also removes all copied models from this update.
        """
        if not self.includes(model_name, record_name):
            return

        if model_name not in models:
            return

        if acc.includes(model_name, record_name):
            return

        # Ensure a record to prevent recursive re-entry
        acc.update_record(model_name, record_name)

        model = models[model_name]

        if destructive:
            record = self.revisions[model_name].pop(record_name)
        else:
            record = self.update_record(model_name, record_name)

        # copy table values into the acc
        for attr_name, attr in model.template.attributes.items():
            if attr_name not in record:
                continue
            val = record[attr_name]

            if attr.link_model_name and attr.attribute_type != "table":
                if isinstance(val, list):
                    for child_id in val:
                        self.sample_revision_tree(attr.link_model_name, child_id, models, acc)
                if isinstance(val, str):
                    self.sample_revision_tree(attr.link_model_name, val, models, acc)
            elif attr.link_model_name and attr.attribute_type == "table":
                if isinstance(val, list):
                    for table_id in val:
                        acc.append_table(
                            model_name,
                            record_name,
                            attr.link_model_name,
                            self.update_record(attr.link_model_name, table_id),
                            attr_name
                        )

    def __len__(self):
        return sum(len(rev) for rev in self.revisions.values())

    def empty(self):
        return len(self) == 0

    def extend(self, other: "UpdateRequest"):
        for model_name, docs in other.revisions.items():
            for record_name, revision in docs.items():
                self.update_record(model_name, record_name, revision)

    # TODO: Add more basic validations?
    def validate(self, models: Dict[str, Model]) -> typing.Iterable[str]:
        for model_name, docs in self.revisions.items():
            if model_name not in models:
                yield f"Model '{model_name}' does not exist"
                continue

            template = models[model_name].template
            for record_name, revision in docs.items():
                for attr, value in revision.items():
                    if attr not in template.attributes:
                        yield f"Model '{model_name}', Attribute '{attr}' does not exist"
                        continue

    def update_record(
        self, model_name: str, record_name: str, attrs: Dict[str, typing.Any] = dict()
    ) -> Dict[str, typing.Any]:
        attrs = {k: normalize_for_magma(v) for k, v in attrs.items()}
        record = self.revisions.setdefault(model_name, {}).setdefault(record_name, {})
        record.update(**attrs)
        return record

    def append_table(
        self,
        parent_model_name: str,
        parent_record_name: str,
        model_name: str,
        attrs: Dict[str, typing.Any],
        attribute_name: str,
    ):
        parent_revision = self.update_record(parent_model_name, parent_record_name, {})
        table = parent_revision.setdefault(attribute_name, [])
        id = f"::{model_name}{len(self.revisions.setdefault(model_name, {})) + 1}"
        table.append(id)
        self.update_record(model_name, id, attrs)
        return id

@serialize
@deserialize
@dataclasses.dataclass
class RetrievalResponse:
    models: Dict[str, Model] = dataclasses.field(default_factory=dict)

    def empty(self):
        for model in self.models.values():
            if not model.empty():
                return False
        return True

    def extend(self, other: "RetrievalResponse"):
        for model_name, model in other.models.items():
            self.models.setdefault(model_name, Model()).extend(model)

@serialize
@deserialize
@dataclasses.dataclass
class QueryRequest:
    query: List[typing.Any] = dataclasses.field(default_factory=list)
    project_name: str = ""
    expand_matrices: Optional[bool] = None
    user_columns: Optional[List[str]] = None
    transpose: Optional[bool] = None
    format: Optional[str] = None

@serialize
@deserialize
@dataclasses.dataclass
class QueryResponse:
    answer: List[typing.Any] = dataclasses.field(default_factory=list)
    format: List[typing.Any] = dataclasses.field(default_factory=list)

class Magma(EtnaClientBase):
    def retrieve(
        self,
        project_name: str,
        model_name="all",
        attribute_names="all",
        record_names: typing.Union[List[str], "all"] = [],
        page: Optional[int] = None,
        page_size: Optional[int] = None,
        order: Optional[str] = None,
        filter: Optional[str] = None,
        hide_templates=True,
        show_disconnected=False,
    ) -> RetrievalResponse:
        args = dict(
            project_name=project_name,
            model_name=model_name,
            attribute_names=attribute_names,
            hide_templates=hide_templates,
            record_names=record_names,
            show_disconnected=show_disconnected,
        )

        if page is None and record_names:
            page = 1
        if page_size is None and record_names:
            page_size = self.batch_size

        if page is not None:
            args["page"] = page
        if page_size is not None:
            args["page_size"] = page_size

        if order is not None:
            args["order"] = order
        if filter is not None:
            args["filter"] = filter

        response = RetrievalResponse(models={})

        def consume_pages():
            while True:
                try:
                    r = self.session.post(self.prepare_url("retrieve"), json=args)
                except HTTPError as e:
                    if b"not found" in e.response.content:
                        if e.response.status_code == 422:
                            break
                    raise e
                paged = from_json(RetrievalResponse, r.content)
                response.extend(paged)

                if "page" not in args:
                    break

                if paged.empty():
                    break
                args["page"] += 1

        if record_names and isinstance(record_names, list):
            for record_name_batch in batch_iterable(record_names, self.batch_size):
                args["record_names"] = record_name_batch
                args["page"] = 1
                consume_pages()
        else:
            args["record_names"] = record_names
            consume_pages()

        return response

    batch_size = 300

    def update(self, update: UpdateRequest):
        response = self.session.post(
            self.prepare_url("update"),
            data=to_json(update),
            headers={"Content-Type": "application/json"},
        )

        return from_json(RetrievalResponse, response.content)

    def query(self, query: QueryRequest):
        response = self.session.post(
            self.prepare_url("query"),
            data=to_json(query),
            headers={"Content-Type": "application/json"},
        )
        if 'format' in query and 'tsv' == query['format']:
            pd_wrapper = partial(pd.read_csv, sep="\t")
            tsv_data = StringIO(response.text)
            return pd_wrapper(tsv_data)

        return from_json(QueryResponse, response.content)

def normalize_for_magma(value):
    # Exhaust generators (such as uploads) before consuming their last yielded value
    if isgenerator(value):
        for value in value:
            pass

    if isinstance(value, File):
        return value.as_magma_file_attribute

    if isinstance(value, Upload):
        magma_file_attribute = value.as_magma_file_attribute
        if magma_file_attribute is None:
            raise ValueError(
                f"Could not determine metis path for upload: '{value.upload_path}', possible bug in etna client."
            )
        return magma_file_attribute

    if isinstance(value, Upload):
        return

    if isinstance(value, list):
        return [normalize_for_magma(v) for v in value]

    return value



import json

import requests
from requests import Session, HTTPError
from requests.auth import AuthBase
from requests.adapters import HTTPAdapter
from typing import Dict, Optional, List, Type
from .auth import TokenAuth
from urllib3 import Retry
from urllib.parse import quote

def encode_path(*segments: str) -> str:
    return "/".join(quote(s) for s in segments)

class EtnaSession(Session):
    DEFAULT_RETRY = Retry(
        total=5,
        backoff_factor=1,
        connect=5,
        read=3,
        status_forcelist=[502, 503, 504],
    )

    @staticmethod
    def assert_status(response: requests.Response, *args, **kwds):
        if 200 <= response.status_code < 300:
            return

        error_message = response.reason

        try:
            err_json = response.json()
            errors = err_json.get("errors", [])
            error = err_json.get("error")

            if error:
                errors.append(error)

            error_message = ", ".join(errors)
            error_message = (
                f"Request failed with {response.status_code}: {error_message}"
            )
        except (TypeError, ValueError, json.JSONDecodeError):
            pass

        raise HTTPError(error_message, response=response)

    def __init__(self, auth: Optional[AuthBase], retry: Optional[Retry] = None):
        super().__init__()
        self.mount(
            "https://",
            HTTPAdapter(
                max_retries=retry if retry is not None else self.DEFAULT_RETRY
            )
        )
        if auth:
            self.auth = auth

        self.hooks['response'].extend([EtnaSession.assert_status])

class EtnaClientBase:
    auth: AuthBase
    session: requests.Session
    hostname: str

    def __init__(self, auth: Optional[AuthBase], hostname: str, session: Optional[Type[requests.Session]] = None, retry: Optional[Retry] = None):
        self.auth = auth
        self.session = (session or EtnaSession)(auth=auth, retry=retry)
        self.hostname = hostname.replace('https://', '')

    @property
    def auth_token(self) -> Optional[bytes]:
        if isinstance(self.session.auth, TokenAuth):
            return self.session.auth.token

    def prepare_url(self, *path: str):
        return f"https://{self.hostname}/{encode_path(*path)}"

    def prepare_url_unsafe(self, path: str):
        return f"https://{self.hostname}/{path}"

    def get_project_scope(self) -> Optional[str]:
        if isinstance(self.session.auth, TokenAuth):
            return self.session.auth.project_scope
        return None

